@@ -2,13 +2,13 @@
 Tests for L{pyflakes.scripts.pyflakes}.
 """
 
+import contextlib
 import os
 import sys
 import shutil
 import subprocess
 import tempfile
 
-from pyflakes.checker import PY2
 from pyflakes.messages import UnusedImport
 from pyflakes.reporter import Reporter
 from pyflakes.api import (
@@ -148,8 +148,7 @@ class TestIterSourceCode(TestCase):
     def makeEmptyFile(self, *parts):
         assert parts
         fpath = os.path.join(self.tempdir, *parts)
-        fd = open(fpath, 'a')
-        fd.close()
+        open(fpath, 'a').close()
         return fpath
 
     def test_emptyDirectory(self):
@@ -180,6 +179,7 @@ class TestIterSourceCode(TestCase):
         """
         os.mkdir(os.path.join(self.tempdir, 'foo'))
         apath = self.makeEmptyFile('foo', 'a.py')
+        self.makeEmptyFile('foo', 'a.py~')
         os.mkdir(os.path.join(self.tempdir, 'bar'))
         bpath = self.makeEmptyFile('bar', 'b.py')
         cpath = self.makeEmptyFile('c.py')
@@ -213,9 +213,35 @@ class TestIterSourceCode(TestCase):
         with open(pythonw, 'w') as fd:
             fd.write('#!/usr/bin/env pythonw\n')
 
+        python3args = os.path.join(self.tempdir, 'g')
+        with open(python3args, 'w') as fd:
+            fd.write('#!/usr/bin/python3 -u\n')
+
+        python2u = os.path.join(self.tempdir, 'h')
+        with open(python2u, 'w') as fd:
+            fd.write('#!/usr/bin/python2u\n')
+
+        python3d = os.path.join(self.tempdir, 'i')
+        with open(python3d, 'w') as fd:
+            fd.write('#!/usr/local/bin/python3d\n')
+
+        python38m = os.path.join(self.tempdir, 'j')
+        with open(python38m, 'w') as fd:
+            fd.write('#! /usr/bin/env python3.8m\n')
+
+        python27 = os.path.join(self.tempdir, 'k')
+        with open(python27, 'w') as fd:
+            fd.write('#!/usr/bin/python2.7   \n')
+
+        # Should NOT be treated as Python source
+        notfirst = os.path.join(self.tempdir, 'l')
+        with open(notfirst, 'w') as fd:
+            fd.write('#!/bin/sh\n#!/usr/bin/python\n')
+
         self.assertEqual(
             sorted(iterSourceCode([self.tempdir])),
-            sorted([python, python2, python3, pythonw]))
+            sorted([python, python2, python3, pythonw, python3args, python2u,
+                    python3d, python38m, python27]))
 
     def test_multipleDirectories(self):
         """
@@ -256,7 +282,9 @@ class TestReporter(TestCase):
         """
         err = StringIO()
         reporter = Reporter(None, err)
-        reporter.syntaxError('foo.py', 'a problem', 3, 7, 'bad line of source')
+        reporter.syntaxError('foo.py', 'a problem', 3,
+                             8 if sys.version_info >= (3, 8) else 7,
+                             'bad line of source')
         self.assertEqual(
             ("foo.py:3:8: a problem\n"
              "bad line of source\n"
@@ -291,10 +319,11 @@ class TestReporter(TestCase):
         reporter = Reporter(None, err)
         reporter.syntaxError('foo.py', 'a problem', 3, len(lines[0]) + 7,
                              '\n'.join(lines))
+        column = 25 if sys.version_info >= (3, 8) else 7
         self.assertEqual(
-            ("foo.py:3:7: a problem\n" +
+            ("foo.py:3:%d: a problem\n" % column +
              lines[-1] + "\n" +
-             "      ^\n"),
+             " " * (column - 1) + "^\n"),
             err.getvalue())
 
     def test_unexpectedError(self):
@@ -323,17 +352,20 @@ class CheckTests(TestCase):
     Tests for L{check} and L{checkPath} which check a file for flakes.
     """
 
+    @contextlib.contextmanager
     def makeTempFile(self, content):
         """
         Make a temporary file containing C{content} and return a path to it.
         """
-        _, fpath = tempfile.mkstemp()
-        if not hasattr(content, 'decode'):
-            content = content.encode('ascii')
-        fd = open(fpath, 'wb')
-        fd.write(content)
-        fd.close()
-        return fpath
+        fd, name = tempfile.mkstemp()
+        try:
+            with os.fdopen(fd, 'wb') as f:
+                if not hasattr(content, 'decode'):
+                    content = content.encode('ascii')
+                f.write(content)
+            yield name
+        finally:
+            os.remove(name)
 
     def assertHasErrors(self, path, errorList):
         """
@@ -371,8 +403,8 @@ class CheckTests(TestCase):
         exception to be raised nor an error indicator to be returned by
         L{check}.
         """
-        fName = self.makeTempFile("def foo():\n\tpass\n\t")
-        self.assertHasErrors(fName, [])
+        with self.makeTempFile("def foo():\n\tpass\n\t") as fName:
+            self.assertHasErrors(fName, [])
 
     def test_checkPathNonExisting(self):
         """
@@ -414,56 +446,56 @@ def baz():
         else:
             self.fail()
 
-        sourcePath = self.makeTempFile(source)
+        with self.makeTempFile(source) as sourcePath:
+            if PYPY:
+                message = 'end of file (EOF) while scanning triple-quoted string literal'
+            else:
+                message = 'invalid syntax'
 
-        if PYPY:
-            message = 'EOF while scanning triple-quoted string literal'
-        else:
-            message = 'invalid syntax'
-
-        self.assertHasErrors(
-            sourcePath,
-            ["""\
-%s:8:11: %s
+            column = 8 if sys.version_info >= (3, 8) else 11
+            self.assertHasErrors(
+                sourcePath,
+                ["""\
+%s:8:%d: %s
     '''quux'''
-          ^
-""" % (sourcePath, message)])
+%s^
+""" % (sourcePath, column, message, ' ' * (column - 1))])
 
     def test_eofSyntaxError(self):
         """
         The error reported for source files which end prematurely causing a
         syntax error reflects the cause for the syntax error.
         """
-        sourcePath = self.makeTempFile("def foo(")
-        if PYPY:
-            result = """\
+        with self.makeTempFile("def foo(") as sourcePath:
+            if PYPY:
+                result = """\
 %s:1:7: parenthesis is never closed
 def foo(
       ^
 """ % (sourcePath,)
-        else:
-            result = """\
+            else:
+                result = """\
 %s:1:9: unexpected EOF while parsing
 def foo(
         ^
 """ % (sourcePath,)
 
-        self.assertHasErrors(
-            sourcePath,
-            [result])
+            self.assertHasErrors(
+                sourcePath,
+                [result])
 
     def test_eofSyntaxErrorWithTab(self):
         """
         The error reported for source files which end prematurely causing a
         syntax error reflects the cause for the syntax error.
         """
-        sourcePath = self.makeTempFile("if True:\n\tfoo =")
-        column = 5 if PYPY else 7
-        last_line = '\t   ^' if PYPY else '\t     ^'
+        with self.makeTempFile("if True:\n\tfoo =") as sourcePath:
+            column = 6 if PYPY else 7
+            last_line = '\t    ^' if PYPY else '\t     ^'
 
-        self.assertHasErrors(
-            sourcePath,
-            ["""\
+            self.assertHasErrors(
+                sourcePath,
+                ["""\
 %s:2:%s: invalid syntax
 \tfoo =
 %s
@@ -479,15 +511,24 @@ def foo(
 def foo(bar=baz, bax):
     pass
 """
-        sourcePath = self.makeTempFile(source)
-        last_line = '       ^\n' if ERROR_HAS_LAST_LINE else ''
-        column = '8:' if ERROR_HAS_COL_NUM else ''
-        self.assertHasErrors(
-            sourcePath,
-            ["""\
+        with self.makeTempFile(source) as sourcePath:
+            if ERROR_HAS_LAST_LINE:
+                if PYPY and sys.version_info >= (3,):
+                    column = 7
+                elif sys.version_info >= (3, 8):
+                    column = 9
+                else:
+                    column = 8
+                last_line = ' ' * (column - 1) + '^\n'
+                columnstr = '%d:' % column
+            else:
+                last_line = columnstr = ''
+            self.assertHasErrors(
+                sourcePath,
+                ["""\
 %s:1:%s non-default argument follows default argument
 def foo(bar=baz, bax):
-%s""" % (sourcePath, column, last_line)])
+%s""" % (sourcePath, columnstr, last_line)])
 
     def test_nonKeywordAfterKeywordSyntaxError(self):
         """
@@ -498,21 +539,30 @@ def foo(bar=baz, bax):
         source = """\
 foo(bar=baz, bax)
 """
-        sourcePath = self.makeTempFile(source)
-        last_line = '            ^\n' if ERROR_HAS_LAST_LINE else ''
-        column = '13:' if ERROR_HAS_COL_NUM or PYPY else ''
+        with self.makeTempFile(source) as sourcePath:
+            if ERROR_HAS_LAST_LINE:
+                if PYPY and sys.version_info >= (3,):
+                    column = 12
+                elif sys.version_info >= (3, 8):
+                    column = 14
+                else:
+                    column = 13
+                last_line = ' ' * (column - 1) + '^\n'
+                columnstr = '%d:' % column
+            else:
+                last_line = columnstr = ''
 
-        if sys.version_info >= (3, 5):
-            message = 'positional argument follows keyword argument'
-        else:
-            message = 'non-keyword arg after keyword arg'
+            if sys.version_info >= (3, 5):
+                message = 'positional argument follows keyword argument'
+            else:
+                message = 'non-keyword arg after keyword arg'
 
-        self.assertHasErrors(
-            sourcePath,
-            ["""\
+            self.assertHasErrors(
+                sourcePath,
+                ["""\
 %s:1:%s %s
 foo(bar=baz, bax)
-%s""" % (sourcePath, column, message, last_line)])
+%s""" % (sourcePath, columnstr, message, last_line)])
 
     def test_invalidEscape(self):
         """
@@ -520,27 +570,32 @@ foo(bar=baz, bax)
         """
         ver = sys.version_info
         # ValueError: invalid \x escape
-        sourcePath = self.makeTempFile(r"foo = '\xyz'")
-        if ver < (3,):
-            decoding_error = "%s: problem decoding source\n" % (sourcePath,)
-        elif PYPY:
-            # pypy3 only
-            decoding_error = """\
-%s:1:6: %s: ('unicodeescape', b'\\\\xyz', 0, 2, 'truncated \\\\xXX escape')
-foo = '\\xyz'
-     ^
-""" % (sourcePath, 'UnicodeDecodeError')
-        else:
-            last_line = '      ^\n' if ERROR_HAS_LAST_LINE else ''
-            # Column has been "fixed" since 3.2.4 and 3.3.1
-            col = 1 if ver >= (3, 3, 1) or ((3, 2, 4) <= ver < (3, 3)) else 2
-            decoding_error = """\
-%s:1:7: (unicode error) 'unicodeescape' codec can't decode bytes \
+        with self.makeTempFile(r"foo = '\xyz'") as sourcePath:
+            if ver < (3,):
+                decoding_error = "%s: problem decoding source\n" % (sourcePath,)
+            else:
+                position_end = 1
+                if PYPY:
+                    column = 6
+                else:
+                    column = 7
+                    # Column has been "fixed" since 3.2.4 and 3.3.1
+                    if ver < (3, 2, 4) or ver[:3] == (3, 3, 0):
+                        position_end = 2
+
+                if ERROR_HAS_LAST_LINE:
+                    last_line = '%s^\n' % (' ' * (column - 1))
+                else:
+                    last_line = ''
+
+                decoding_error = """\
+%s:1:%d: (unicode error) 'unicodeescape' codec can't decode bytes \
 in position 0-%d: truncated \\xXX escape
 foo = '\\xyz'
-%s""" % (sourcePath, col, last_line)
-        self.assertHasErrors(
-            sourcePath, [decoding_error])
+%s""" % (sourcePath, column, position_end, last_line)
+
+            self.assertHasErrors(
+                sourcePath, [decoding_error])
 
     @skipIf(sys.platform == 'win32', 'unsupported on Windows')
     def test_permissionDenied(self):
@@ -551,24 +606,24 @@ foo = '\\xyz'
         if os.getuid() == 0:
             self.skipTest('root user can access all files regardless of '
                           'permissions')
-        sourcePath = self.makeTempFile('')
-        os.chmod(sourcePath, 0)
-        count, errors = self.getErrors(sourcePath)
-        self.assertEqual(count, 1)
-        self.assertEqual(
-            errors,
-            [('unexpectedError', sourcePath, "Permission denied")])
+        with self.makeTempFile('') as sourcePath:
+            os.chmod(sourcePath, 0)
+            count, errors = self.getErrors(sourcePath)
+            self.assertEqual(count, 1)
+            self.assertEqual(
+                errors,
+                [('unexpectedError', sourcePath, "Permission denied")])
 
     def test_pyflakesWarning(self):
         """
         If the source file has a pyflakes warning, this is reported as a
         'flake'.
         """
-        sourcePath = self.makeTempFile("import foo")
-        count, errors = self.getErrors(sourcePath)
-        self.assertEqual(count, 1)
-        self.assertEqual(
-            errors, [('flake', str(UnusedImport(sourcePath, Node(1), 'foo')))])
+        with self.makeTempFile("import foo") as sourcePath:
+            count, errors = self.getErrors(sourcePath)
+            self.assertEqual(count, 1)
+            self.assertEqual(
+                errors, [('flake', str(UnusedImport(sourcePath, Node(1), 'foo')))])
 
     def test_encodedFileUTF8(self):
         """
@@ -579,15 +634,15 @@ foo = '\\xyz'
 # coding: utf-8
 x = "%s"
 """ % SNOWMAN).encode('utf-8')
-        sourcePath = self.makeTempFile(source)
-        self.assertHasErrors(sourcePath, [])
+        with self.makeTempFile(source) as sourcePath:
+            self.assertHasErrors(sourcePath, [])
 
     def test_CRLFLineEndings(self):
         """
         Source files with Windows CR LF line endings are parsed successfully.
         """
-        sourcePath = self.makeTempFile("x = 42\r\n")
-        self.assertHasErrors(sourcePath, [])
+        with self.makeTempFile("x = 42\r\n") as sourcePath:
+            self.assertHasErrors(sourcePath, [])
 
     def test_misencodedFileUTF8(self):
         """
@@ -599,22 +654,21 @@ x = "%s"
 # coding: ascii
 x = "%s"
 """ % SNOWMAN).encode('utf-8')
-        sourcePath = self.makeTempFile(source)
-
-        if PYPY and sys.version_info < (3, ):
-            message = ('\'ascii\' codec can\'t decode byte 0xe2 '
-                       'in position 21: ordinal not in range(128)')
-            result = """\
+        with self.makeTempFile(source) as sourcePath:
+            if PYPY and sys.version_info < (3, ):
+                message = ('\'ascii\' codec can\'t decode byte 0xe2 '
+                           'in position 21: ordinal not in range(128)')
+                result = """\
 %s:0:0: %s
 x = "\xe2\x98\x83"
         ^\n""" % (sourcePath, message)
 
-        else:
-            message = 'problem decoding source'
-            result = "%s: problem decoding source\n" % (sourcePath,)
+            else:
+                message = 'problem decoding source'
+                result = "%s: problem decoding source\n" % (sourcePath,)
 
-        self.assertHasErrors(
-            sourcePath, [result])
+            self.assertHasErrors(
+                sourcePath, [result])
 
     def test_misencodedFileUTF16(self):
         """
@@ -626,9 +680,9 @@ x = "\xe2\x98\x83"
 # coding: ascii
 x = "%s"
 """ % SNOWMAN).encode('utf-16')
-        sourcePath = self.makeTempFile(source)
-        self.assertHasErrors(
-            sourcePath, ["%s: problem decoding source\n" % (sourcePath,)])
+        with self.makeTempFile(source) as sourcePath:
+            self.assertHasErrors(
+                sourcePath, ["%s: problem decoding source\n" % (sourcePath,)])
 
     def test_checkRecursive(self):
         """
@@ -636,24 +690,25 @@ x = "%s"
         and reporting problems.
         """
         tempdir = tempfile.mkdtemp()
-        os.mkdir(os.path.join(tempdir, 'foo'))
-        file1 = os.path.join(tempdir, 'foo', 'bar.py')
-        fd = open(file1, 'wb')
-        fd.write("import baz\n".encode('ascii'))
-        fd.close()
-        file2 = os.path.join(tempdir, 'baz.py')
-        fd = open(file2, 'wb')
-        fd.write("import contraband".encode('ascii'))
-        fd.close()
-        log = []
-        reporter = LoggingReporter(log)
-        warnings = checkRecursive([tempdir], reporter)
-        self.assertEqual(warnings, 2)
-        self.assertEqual(
-            sorted(log),
-            sorted([('flake', str(UnusedImport(file1, Node(1), 'baz'))),
-                    ('flake',
-                     str(UnusedImport(file2, Node(1), 'contraband')))]))
+        try:
+            os.mkdir(os.path.join(tempdir, 'foo'))
+            file1 = os.path.join(tempdir, 'foo', 'bar.py')
+            with open(file1, 'wb') as fd:
+                fd.write("import baz\n".encode('ascii'))
+            file2 = os.path.join(tempdir, 'baz.py')
+            with open(file2, 'wb') as fd:
+                fd.write("import contraband".encode('ascii'))
+            log = []
+            reporter = LoggingReporter(log)
+            warnings = checkRecursive([tempdir], reporter)
+            self.assertEqual(warnings, 2)
+            self.assertEqual(
+                sorted(log),
+                sorted([('flake', str(UnusedImport(file1, Node(1), 'baz'))),
+                        ('flake',
+                         str(UnusedImport(file2, Node(1), 'contraband')))]))
+        finally:
+            shutil.rmtree(tempdir)
 
 
 class IntegrationTests(TestCase):
@@ -661,6 +716,12 @@ class IntegrationTests(TestCase):
     Tests of the pyflakes script that actually spawn the script.
     """
 
+    # https://bitbucket.org/pypy/pypy/issues/3069/pypy36-on-windows-incorrect-line-separator
+    if PYPY and sys.version_info >= (3,) and WIN:
+        LINESEP = '\n'
+    else:
+        LINESEP = os.linesep
+
     def setUp(self):
         self.tempdir = tempfile.mkdtemp()
         self.tempfilepath = os.path.join(self.tempdir, 'temp')
@@ -701,9 +762,6 @@ class IntegrationTests(TestCase):
         if sys.version_info >= (3,):
             stdout = stdout.decode('utf-8')
             stderr = stderr.decode('utf-8')
-        # Workaround https://bitbucket.org/pypy/pypy/issues/2350
-        if PYPY and PY2 and WIN:
-            stderr = stderr.replace('\r\r\n', '\r\n')
         return (stdout, stderr, rv)
 
     def test_goodFile(self):
@@ -711,8 +769,7 @@ class IntegrationTests(TestCase):
         When a Python source file is all good, the return code is zero and no
         messages are printed to either stdout or stderr.
         """
-        fd = open(self.tempfilepath, 'a')
-        fd.close()
+        open(self.tempfilepath, 'a').close()
         d = self.runPyflakes([self.tempfilepath])
         self.assertEqual(d, ('', '', 0))
 
@@ -721,12 +778,11 @@ class IntegrationTests(TestCase):
         When a Python source file has warnings, the return code is non-zero
         and the warnings are printed to stdout.
         """
-        fd = open(self.tempfilepath, 'wb')
-        fd.write("import contraband\n".encode('ascii'))
-        fd.close()
+        with open(self.tempfilepath, 'wb') as fd:
+            fd.write("import contraband\n".encode('ascii'))
         d = self.runPyflakes([self.tempfilepath])
         expected = UnusedImport(self.tempfilepath, Node(1), 'contraband')
-        self.assertEqual(d, ("%s%s" % (expected, os.linesep), '', 1))
+        self.assertEqual(d, ("%s%s" % (expected, self.LINESEP), '', 1))
 
     def test_errors_io(self):
         """
@@ -736,7 +792,7 @@ class IntegrationTests(TestCase):
         """
         d = self.runPyflakes([self.tempfilepath])
         error_msg = '%s: No such file or directory%s' % (self.tempfilepath,
-                                                         os.linesep)
+                                                         self.LINESEP)
         self.assertEqual(d, ('', error_msg, 1))
 
     def test_errors_syntax(self):
@@ -745,13 +801,12 @@ class IntegrationTests(TestCase):
         exist, say), then the return code is non-zero and the errors are
         printed to stderr.
         """
-        fd = open(self.tempfilepath, 'wb')
-        fd.write("import".encode('ascii'))
-        fd.close()
+        with open(self.tempfilepath, 'wb') as fd:
+            fd.write("import".encode('ascii'))
         d = self.runPyflakes([self.tempfilepath])
-        error_msg = '{0}:1:{2}: invalid syntax{1}import{1}    {3}^{1}'.format(
-            self.tempfilepath, os.linesep, 5 if PYPY else 7, '' if PYPY else '  ')
-        self.assertEqual(d, ('', error_msg, True))
+        error_msg = '{0}:1:{2}: invalid syntax{1}import{1}     {3}^{1}'.format(
+            self.tempfilepath, self.LINESEP, 6 if PYPY else 7, '' if PYPY else ' ')
+        self.assertEqual(d, ('', error_msg, 1))
 
     def test_readFromStdin(self):
         """
@@ -759,19 +814,22 @@ class IntegrationTests(TestCase):
         """
         d = self.runPyflakes([], stdin='import contraband')
         expected = UnusedImport('<stdin>', Node(1), 'contraband')
-        self.assertEqual(d, ("%s%s" % (expected, os.linesep), '', 1))
+        self.assertEqual(d, ("%s%s" % (expected, self.LINESEP), '', 1))
 
 
 class TestMain(IntegrationTests):
     """
     Tests of the pyflakes main function.
     """
+    LINESEP = os.linesep
 
     def runPyflakes(self, paths, stdin=None):
         try:
             with SysStreamCapturing(stdin) as capture:
                 main(args=paths)
         except SystemExit as e:
-            return (capture.output, capture.error, e.code)
+            self.assertIsInstance(e.code, bool)
+            rv = int(e.code)
+            return (capture.output, capture.error, rv)
         else:
             raise RuntimeError('SystemExit not raised')
